#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `wspr` package."""
import sys
import unittest

from wspr_varint.varint import VarInt


class VarIntTest(unittest.TestCase):
    """"""

    def setUp(self):
        """Set up test fixtures, if any."""
        integers = [
            -4,
            -3,
            -2,
            -1,
        ]
        self.test_cases_from_int = [
            (integer, VarInt.from_int(integer)) for integer in integers
        ]

    def tearDown(self):
        """Tear down test fixtures, if any."""
        pass

    def test_conversion_from_to(self):
        for integer in range(0, 4294967296 + 1000000, 33333):
            expected = VarInt.from_int(integer)
            calculated = VarInt.from_bytes(expected.to_bytes())
            self.assertEqual(calculated.to_int(), expected.to_int())

    def test_int_to_bytes(self):
        """"""
        # 111111xx: integers between and including -1 to -4
        cases = [
            (-1, 0b11111100),
            (-2, 0b11111101),
            (-3, 0b11111110),
            (-4, 0b11111111),
        ]
        for integer, binary in cases:
            varint = VarInt.from_int(integer)
            self.assertEqual(binary_to_bytes(binary), varint.to_bytes())

        # 111110__ + varint: Negative recursive varint
        # TODO

        # 0xxxxxxx: 7-bit positive number
        for integer in range(0, 128):
            varint = VarInt.from_int(integer)
            # self.assertEqual(integer & 0b10000000, 0b00000000)
            self.assertEqual(binary_to_bytes(integer), varint.to_bytes())

        # 10xxxxxx + 1 byte: 14-bit positive number
        for integer in range(128, 16384):
            varint = VarInt.from_int(integer)
            # self.assertEqual(integer & 0b1100000000000000, 0b1000000000000000)
            self.assertEqual(binary_to_bytes(integer | 0b1000000000000000), varint.to_bytes())

        # 110xxxxx + 2 bytes: 21-bit positive number
        for integer in range(16384, 16384 + 5000):
            varint = VarInt.from_int(integer)
            # self.assertEqual(integer & 0b111000000000000000000000, 0b110000000000000000000000)
            self.assertEqual(binary_to_bytes(integer | 0b110000000000000000000000), varint.to_bytes())
        for integer in range(2097152 - 5000, 2097152):
            varint = VarInt.from_int(integer)
            # self.assertEqual(integer & 0b1110000000000000, 0b1100000000000000)
            self.assertEqual(binary_to_bytes(integer | 0b110000000000000000000000), varint.to_bytes())

        # 1110xxxx + 3 bytes: 28-bit positive number
        for integer in range(2097152, 2097152 + 5000):
            varint = VarInt.from_int(integer)
            # self.assertEqual(integer & 0b11110000000000000000000000000000, 0b11100000000000000000000000000000)
            self.assertEqual(binary_to_bytes(integer | 0b11100000000000000000000000000000), varint.to_bytes())
        for integer in range(268435456 - 5000, 268435456):
            varint = VarInt.from_int(integer)
            # self.assertEqual(integer & 0b11110000000000000000000000000000, 0b11100000000000000000000000000000)
            self.assertEqual(binary_to_bytes(integer | 0b11100000000000000000000000000000), varint.to_bytes())
        self.assertEqual(binary_to_bytes(0b11101111111111111111111111111111), VarInt(268435455).to_bytes())

        # 111100__ + int (32-bit): 32-bit positive number
        for integer in range(268435456, 268435456 + 5000):
            varint = VarInt.from_int(integer)
            # self.assertEqual(integer & 0b11110000000000000000000000000000, 0b11100000000000000000000000000000)
            self.assertEqual(binary_to_bytes(integer | 0b1111000000000000000000000000000000000000), varint.to_bytes())
        for integer in range(4294967296 - 5000, 4294967296):
            varint = VarInt.from_int(integer)
            # self.assertEqual(integer & 0b11110000000000000000000000000000, 0b11100000000000000000000000000000)
            self.assertEqual(binary_to_bytes(integer | 0b1111000000000000000000000000000000000000), varint.to_bytes())
        self.assertEqual(binary_to_bytes(0b1111000011111111111111111111111111111111), VarInt(4294967295).to_bytes())

        # 111101__ + long (64-bit): 64-bit number
        for integer in range(4294967296, 4294967296 + 5000):
            varint = VarInt.from_int(integer)
            # self.assertEqual(integer & 0b11110000000000000000000000000000, 0b11100000000000000000000000000000)
            self.assertEqual(
                binary_to_bytes(integer | 0b111101000000000000000000000000000000000000000000000000000000000000000000),
                varint.to_bytes())
        # for integer in range(2 ^ 64 - 5000, 2 ^ 64):
        for integer in range(sys.maxsize // 2 - 5000, sys.maxsize // 2):
            varint = VarInt.from_int(integer)
            # self.assertEqual(integer & 0b11110000000000000000000000000000, 0b11100000000000000000000000000000)
            self.assertEqual(
                binary_to_bytes(integer | 0b111101000000000000000000000000000000000000000000000000000000000000000000),
                varint.to_bytes())
        # 9223372036854775807 // 2
        self.assertEqual(binary_to_bytes(
            4611686018427387903 | 0b111101000000000000000000000000000000000000000000000000000000000000000000),
                         VarInt(sys.maxsize // 2).to_bytes())


def binary_to_bytes(binary: bin) -> bytes:
    """Convert a binary format to bytes in the big endian (network) format."""
    return binary.to_bytes(max(binary.bit_length() // 8, 1), 'big')



from __future__ import annotations
import struct


class VarInt:
    """
    Variable length integer as specified in the mumble documentation.
    https://mumble-protocol.readthedocs.io/en/latest/voice_data.html#variable-length-integer-encoding
    """

    def __init__(self, integer):
        """"""
        self.integer = integer

    @classmethod
    def from_int(cls, integer: int) -> VarInt:
        """"""
        return cls(integer)

    @classmethod
    def from_bytes(cls, bytes_in: bytes) -> VarInt:
        """"""
        length = len(bytes_in)
        if length == 0:
            raise ValueError("Zero length VarInt")
        byte0, = struct.unpack("!B", bytes_in[0:1])
        # 111110__ + varint: Negative recursive varint
        # Throw __ away and append varint of positive integer
        if byte0 & 0b11111100 == 0b11111000:
            result = cls.from_bytes(bytes_in[1:])
            result.integer = -result.integer
            return result
        if length == 1:
            # 0xxxxxxx: 7-bit positive number
            # 01111111 -> 127
            if byte0 & 0b10000000 == 0b00000000:
                return cls(byte0)
            # 111111xx: Byte-inverted negative two bit number (~xx)
            # values between and including -1 and -4
            if byte0 & 0b11111100 == 0b11111100:
                return cls(~(byte0 & 0b00000011))
        elif length == 2:
            # 10xxxxxx + 1 byte: 14-bit positive number
            # 00111111 11111111 -> 16383
            if byte0 & 0b11000000 == 0b10000000:
                first, = struct.unpack("!H", bytes_in)
                return cls(first & 0b0011111111111111)
        elif length == 3:
            # 110xxxxx + 2 bytes: 21-bit positive number
            # 00011111 11111111 11111111 -> 2097151
            if byte0 & 0b11100000 == 0b11000000:
                first, second = struct.unpack("!BH", bytes_in)
                return cls(((first & 0b00011111) << 16) + second)
        elif length == 4:
            # 1110xxxx + 3 bytes: 28-bit positive number
            # 00001111 11111111 11111111 11111111 -> 268435455
            if byte0 & 0b11110000 == 0b11100000:
                first, = struct.unpack("!L", bytes_in)
                return cls(first & 0b00001111111111111111111111111111)
        elif length == 5:
            # 111100__ + int (32-bit): 32-bit positive number
            # 00000000 11111111 11111111 11111111 11111111 -> 4294967295
            if byte0 & 0b11111100 == 0b11110000:
                first, second = struct.unpack("!BL", bytes_in)
                return cls(second)
        elif length == 9:
            # 111101__ + long (64-bit): 64-bit number
            if byte0 & 0b11111100 == 0b11110100:
                first, second = struct.unpack("!BQ", bytes_in)
                return cls(second)
        return cls(0)

    def to_int(self) -> int:
        """"""
        return self.integer

    @property
    def int(self) -> int:
        """"""
        return self.to_int()

    def to_bytes(self) -> bytes:
        """"""
        return self.__to_bytes(self.integer)

    @property
    def bytes(self) -> bytes:
        """"""
        return self.to_bytes()

    def __to_bytes(self, integer) -> bytes:
        """"""
        # 111111xx: Byte-inverted negative two bit number (~xx)
        # values between and including -1 and -4
        # where xx ->
        #   00: -1
        #   01: -2
        #   10: -3
        #   11: -4
        if -5 < integer < 0:
            return struct.pack("!B", (0b11111100 | ~integer))
        # 111110__ + varint: Negative recursive varint
        # Ignore __ and append varint of positive integer
        if integer < 0:
            return struct.pack("!B", 0b11111000) + self.__to_bytes(-integer)
        # 0xxxxxxx: 7-bit positive number
        # 01111111 -> 127
        if integer < 128:
            return struct.pack("!B", integer)
        # 10xxxxxx + 1 byte: 14-bit positive number
        # 00111111 11111111 -> 16383
        if integer < 16384:
            return struct.pack("!H", 0b1000000000000000 | integer)
        # 110xxxxx + 2 bytes: 21-bit positive number
        # 00011111 11111111 11111111 -> 2097151
        if integer < 2097152:
            return struct.pack("!BH", 0b11000000 | (integer >> 16), integer & 0b1111111111111111)
        # 1110xxxx + 3 bytes: 28-bit positive number
        # 00001111 11111111 11111111 11111111 -> 268435455
        if integer < 268435456:
            return struct.pack("!L", 0b11100000000000000000000000000000 | integer)
        # 111100__ + int (32-bit): 32-bit positive number
        # 00000000 11111111 11111111 11111111 11111111 -> 4294967295
        if integer < 4294967296:
            return struct.pack("!BL", 0b11110000, integer)
        # 111101__ + long (64-bit): 64-bit number
        return struct.pack("!BQ", 0b11110100, integer)

    def __add__(self, other: [VarInt, int, bytes]) -> VarInt:
        """"""
        if isinstance(other, VarInt):
            return VarInt.from_int(self.int + other.int)
        elif isinstance(other, int):
            return VarInt.from_int(self.int + other)
        elif isinstance(other, bytes):
            return self + VarInt.from_bytes(other)
        raise TypeError(f"Unsupported operand type(s) for +: '{type(self)}' and '{type(other)}'")

    def __mul__(self, other: [VarInt, int, bytes]) -> VarInt:
        """"""
        if isinstance(other, VarInt):
            return VarInt.from_int(self.int * other.int)
        elif isinstance(other, int):
            return VarInt.from_int(self.int * other)
        elif isinstance(other, bytes):
            return self * VarInt.from_bytes(other)
        raise TypeError(f"Unsupported operand type(s) for *: '{type(self)}' and '{type(other)}'")

    def __repr__(self) -> repr:
        """"""
        return f"{self.integer} -> {self.to_bytes()}"



from setuptools import setup, find_packages
import pathlib


def pin_major(requirement: str) -> str:
    """Return a dependency pinned to the current major version."""
    library_name, library_version = requirement.split("==")
    major, *_ = library_version.split(".")
    return f"{library_name}>={int(major)}, <{int(major) + 1}"


with pathlib.Path("README.rst").open() as readme_file:
    readme = readme_file.read()

try:
    with pathlib.Path("requirements.txt").open() as f:
        lines = f.read().splitlines()
        requirements = filter(lambda line: not line.startswith("#") and len(line) > 0, lines)
        requirements = [pin_major(requirement) for requirement in requirements]
except FileNotFoundError:
    requirements = []

try:
    with pathlib.Path("requirements_dev.txt").open() as f:
        lines = f.read().splitlines()
        requirements_dev = filter(lambda line: not line.startswith("#") and len(line) > 0, lines)
        requirements_dev = [pin_major(requirement) for requirement in requirements_dev]
except FileNotFoundError:
    requirements_dev = []

with pathlib.Path("HISTORY.rst").open() as history_file:
    history = history_file.read()

setup_requirements = []

test_requirements = []

classifiers = [
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Natural Language :: English",
    "Programming Language :: Python :: 3.8",
]

# https://setuptools.readthedocs.io/en/latest/setuptools.html#including-data-files
# package_data = {
#     "": ["*.txt", "*.rst"],
# }

setup(
    author="Arkadiusz Michał Ryś",
    author_email="Arkadiusz.Michal.Rys@gmail.com",
    classifiers=classifiers,
    description="A variable length integer implementation for the wspr project.",
    include_package_data=True,
    install_requires=requirements,
    keywords="wspr_varint",
    license="MIT license",
    long_description=readme + "\n\n" + history,
    name="wspr_varint",
    # package_data=package_data,
    packages=find_packages(include=["wspr_varint"]),
    python_requires=">=3.8, <4",
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="",
    version="1.1.1",
    zip_safe=False,
)
